
#  ============LICENSE_START===============================================
#  Copyright (C) 2020 Nordix Foundation. All rights reserved.
#  ========================================================================
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ============LICENSE_END=================================================
#

from flask import Flask, request, Response
from flask import send_file
from time import sleep
import time
from datetime import datetime
import json
import traceback
import logging
import socket
import subprocess
import os
import shutil
import sys

app = Flask(__name__)

# list of callback messages
msg_callbacks={}

# Server info
HOST_IP = "::"
HOST_PORT = 2222



NOTINSTALLED="NOT-INSTALLED"
INSTALLED="INSTALLED"
NONE="NONE"
UNDEFINED_NAMESPACE="#NONE#"
STATUS_JSON="chart_status.json"

BASE_DIR=None

if len(sys.argv) > 1:
    if isinstance(sys.argv[1], str):
        BASE_DIR = sys.argv[1]

if BASE_DIR is None and os.path.isdir("/var"):
    BASE_DIR="/var"
    try:
        os.chdir(BASE_DIR)
    except OSError:
       print("Cannot change dir to "+BASE_DIR)

if BASE_DIR is None:
    print("BASE_DIR cannot be set")
    exit(1)

print("base dir: " + str(BASE_DIR))

CHART_REPO=BASE_DIR+"/rappman"

#I'm alive function
@app.route('/',
    methods=['GET'])
def index():
    return "rappman OK", 200

@app.route('/charts',
    methods=['GET'])
def get_charts():

    files=[]
    for f in os.listdir(CHART_REPO):
        files.append(f)
    return json.dumps(files), 200

@app.route('/charts/<string:chartname>',
    methods=['PUT'])
def put_upload_chart(chartname):

    print("put_upload_chart Chartname: "+chartname)

    for f in request.files:
        print(str(f))

    chart_status=None


    chart_dir=CHART_REPO+"/"+chartname
    status_file=chart_dir+"/"+STATUS_JSON
    if not os.path.isdir(chart_dir):
        try:
            os.mkdir(chart_dir)
        except OSError:
            return "Cannot create chart dir "+chart_dir,400

    if os.path.isfile(status_file):
        with open(status_file, 'r') as infile:
            chart_status=json.load(infile)
    else:
        chart_status={}
        chart_status["name"]=chartname
        chart_status["install_state"]=NOTINSTALLED
        chart_status["install_namespace"]=UNDEFINED_NAMESPACE
        chart_status["chart_file_name"]=NONE
        chart_status["chart_file_status"]=NOTINSTALLED
        chart_status["override_file_name"]=NONE
        chart_status["override_file_status"]=NOTINSTALLED
        with open(status_file, 'w') as outfile:
            json.dump(chart_status, outfile)

    file_count=0
    if 'chart' in request.files:
        file=request.files['chart']
        if not file.filename.endswith(".tgz"):
            return "Chart file has incorrect file extension, expected 'tgz'",400
        chart_file_name=chart_dir+"/"+file.filename
        file.save(chart_file_name)
        file_count=file_count+1
        chart_status["chart_file_status"]=NOTINSTALLED
        chart_status["chart_file_name"]=file.filename

    if 'values' in request.files:
        file=request.files['values']
        if not file.filename.endswith(".yaml") and not file.filename.endswith(".yml"):
            return "Override file has incorrect file extension, expected 'yaml' or 'yml'",400
        chart_file_name=chart_dir+"/"+file.filename
        file.save(chart_file_name)
        file_count=file_count+1
        chart_status["override_file_status"]=NOTINSTALLED
        chart_status["override_file_name"]=file.filename

    if file_count==0:
        return "Nor chart or override file in request",400

    if file_count != len(request.files):
        return "Unknown file(s) sent in request",400

    with open(status_file, 'w') as outfile:
        json.dump(chart_status, outfile)

    return "OK", 200

@app.route('/charts/<string:chartname>',
    methods=['POST'])
def post_action_chart(chartname):

    print("post_action_chart Chartname: "+chartname)

    chart_dir=CHART_REPO+"/"+chartname

    action_id=request.args.get('action')
    if (action_id is None):
        return "Action id missing", 400

    if action_id != "install" and action_id != "uninstall":
        return "Action id should be install or uninstall", 400

    if not os.path.isdir(chart_dir):
        return "No chart with name: " + chartname, 404

    status_file=chart_dir+"/"+STATUS_JSON
    try:
        with open(status_file, 'r') as infile:
            chart_status=json.load(infile)
    except Exception as err:
        return "Internal error - "+str(err), 500
    os.chdir(chart_dir)

    print("curdir: " + str(os.getcwd()))
    if action_id == "install":

        namespace=request.args.get("namespace")
        if namespace is None:
            namespace=chart_status["install_namespace"]
            if namespace == UNDEFINED_NAMESPACE:
                return "Namespace missing", 400
        else:
            if len(namespace) == 0:
                return "Namespace is zero lengths", 400
            if chart_status["install_namespace"] == UNDEFINED_NAMESPACE:
                chart_status["install_namespace"]=namespace
            elif chart_status["install_state"]==NOTINSTALLED:
                chart_status["install_namespace"]=namespace
            elif chart_status["install_namespace"] != namespace and chart_status["install_state"]==INSTALLED:
                return "Cannot change namespace for installed chart", 400

        cmd_arr=["kubectl", "get", "namespace", namespace]
        print("cmd arr: "+str(cmd_arr))
        result = subprocess.run(cmd_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        print(str(result.stderr))
        print(str(result.stdout))

        if result.returncode != 0:
            cmd_arr=["kubectl", "create", "namespace", namespace]
            print("cmd arr: "+str(cmd_arr))
            result = subprocess.run(cmd_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            print(str(result.stderr))
            print(str(result.stdout))
            if result.returncode != 0:
                return "Cannot create namespace: "+namespace, 500

        upgrade=False
        cmd_arr=[]
        cmd_arr.append("helm")
        if chart_status["install_state"]==NOTINSTALLED:
            cmd_arr.append("install")
        else:
            cmd_arr.append("upgrade")
            upgrade=True

        cmd_arr.append("-n")
        cmd_arr.append(namespace)

        cmd_arr.append(chartname)
        if chart_status["chart_file_name"]==NONE:
            return "Chart is missing", 500

        chart_status["install_state"]=INSTALLED
        chart_status["chart_file_status"]=INSTALLED
        cmd_arr.append(chart_status["chart_file_name"])
        if chart_status["override_file_name"]!=NONE:
            chart_status["override_file_status"]=INSTALLED
            cmd_arr.append("-f")
            cmd_arr.append(chart_status["override_file_name"])

        print("cmd arr: "+str(cmd_arr))
        result = subprocess.run(cmd_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        print(str(result.stderr))
        print(str(result.stdout))

        res_str="Installed"
        if upgrade == True:
            res_str="Upgraded"
        res_code=200
        if result.returncode != 0:
            res_str="Installation failed"
            if upgrade == True:
                res_str="Upgrade failed"
            res_code=400
        else:
            with open(status_file, 'w') as outfile:
                json.dump(chart_status, outfile)

        return res_str,res_code

    else:
        if chart_status["install_state"]==NOTINSTALLED:
            return "Chart not installed",400

        cmd_arr=[]
        cmd_arr.append("helm")
        cmd_arr.append("uninstall")
        cmd_arr.append("-n")
        cmd_arr.append(chart_status["install_namespace"])
        cmd_arr.append(chartname)

        print("cmd arr: "+str(cmd_arr))
        result = subprocess.run(cmd_arr, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        print(str(result.stderr))
        print(str(result.stdout))

        res_str="Uninstalled"
        res_code=200
        if result.returncode != 0:
            res_str="Uninstallation failed"
            res_code=400
        else:
            chart_status["install_state"]=NOTINSTALLED
            chart_status["chart_file_status"]=NOTINSTALLED
            if chart_status["override_file_name"]!=NONE:
                chart_status["override_file_status"]=NOTINSTALLED
            with open(status_file, 'w') as outfile:
                json.dump(chart_status, outfile)

        return res_str,res_code


@app.route('/charts/<string:chartname>',
    methods=['DELETE'])
def delete_uninstall_chart(chartname):

    print("put_install_chart Chartname: "+chartname)

    chart_dir=CHART_REPO+"/"+chartname

    if not os.path.isdir(chart_dir):
        return "No chart with name: " + chartname, 404

    status_file=chart_dir+"/"+STATUS_JSON
    try:
        with open(status_file, 'r') as infile:
            chart_status=json.load(infile)
    except Exception as err:
        return "Internal error - "+str(err), 500

    if chart_status["install_state"]!=NOTINSTALLED:
        return "Cannot delete, chart: " + chartname + " is still installed", 400
    try:
        shutil.rmtree(chart_dir)
    except OSError as e:
        print("Error: %s : %s" % (chart_dir, e.strerror))
        return "Delete chart failed", 500

    return "OK",200


@app.route('/charts/<string:chartname>/status',
    methods=['GET'])
def get_chart_status(chartname):

    print("get_chart_status Chartname: "+chartname)


    chart_dir=CHART_REPO+"/"+chartname

    if not os.path.isdir(chart_dir):
        return "No chart with name: " + chartname, 404

    status_file=chart_dir+"/"+STATUS_JSON
    try:
        with open(status_file, 'r') as infile:
            chart_status=json.load(infile)
            return json.dumps(chart_status),200
    except Exception as err:
        return "Internal error - " + str(err), 500

@app.route('/charts/<string:chartname>/chart',
    methods=['GET'])
def get_chart(chartname):

    print("get_chart Chartname: "+chartname)

    chart_dir=CHART_REPO+"/"+chartname

    if not os.path.isdir(chart_dir):
        return "No chart with name: " + chartname, 404

    status_file=chart_dir+"/"+STATUS_JSON
    try:
        with open(status_file, 'r') as outfile:
            chart_status=json.load(outfile)
    except Exception as err:
        return "Internal error - " + str(err), 500

    if chart_status["chart_file_name"] == NONE:
        return "File not found", 404
    chart_file=chart_dir+"/"+chart_status["chart_file_name"]

    try:
        return send_file(chart_file, attachment_filename=chart_status["chart_file_name"])
    except Exception as err:
        return "Internal error - " + str(err), 500

@app.route('/charts/<string:chartname>/override',
    methods=['GET'])
def get_override(chartname):

    print("get_chart Chartname: "+chartname)

    chart_dir=CHART_REPO+"/"+chartname

    if not os.path.isdir(chart_dir):
        return "No chart with name: " + chartname, 404

    status_file=chart_dir+"/"+STATUS_JSON
    try:
        with open(status_file, 'r') as outfile:
            chart_status=json.load(outfile)
    except Exception as err:
        return "Internal error - " + str(err), 500

    if chart_status["override_file_name"] == NONE:
        return "File not found", 404
    ovr_file=chart_dir+"/"+chart_status["override_file_name"]

    try:
        return send_file(ovr_file, attachment_filename=chart_status["override_file_name"])
    except Exception as err:
        return "Internal error - " + str(err), 500


### Main function ###

if __name__ == "__main__":
    app.run(port=HOST_PORT, host=HOST_IP)
